/*
 * A basic video and keyboard driver for MoDapple.
 * 
 * * mod_type should be 0x54545950 (TTYP)
 * * mod_name should be the display name of the module
 * * mod_pref should be the name prefixed to module functions
 */

#include <curses.h>
#include <stdint.h>
#include <windows.h>

__declspec(dllexport) uint32_t mod_type=0x50595454;
__declspec(dllexport) uint8_t *mod_name="Curses";
__declspec(dllexport) uint8_t *mod_pref="ttycurses";
__declspec(dllexport) uint8_t *mod_vers="0.04";
__declspec(dllexport) uint8_t *mod_copr="Copyright 2017 Steve Nickolas";

typedef struct {
 uint8_t iiemode; /* emulating a //e */
 uint8_t store80,
         rdaux,
         wraux,
         lcaux,
         cxoverride,
         c3override,
         txt80,
         dblgrx,
         mostxt,
         iou,
         an3,
         alt[2];
 uint8_t lcread,
         lcwrite,
         lcalt;
 uint8_t nogx,
         split,
         page2,
         hires;

 uint8_t *ram;
 uint8_t enh;
} IIE;

static IIE *iie;

static uint8_t *vramtp[4];

static int nogx, split, page2, hires, ttyok;
__declspec(dllexport) uint8_t ttycurses_ttykey;

static uint32_t mtapprox[32]={
 'O', 'A', '^', 0xE8, 0xFB, 0xFB, 0xF0, ' ',
 ACS_LARROW, '_', ACS_DARROW, ACS_UARROW, 0xDF, '<', 0xFE, ACS_LARROW,
 ACS_RARROW, ACS_DARROW, ACS_UARROW, ACS_HLINE, 0xC0, ACS_RARROW, 0xAA, 0xAA,
 '[', ']', 0xDE, ACS_DIAMOND, 0xF0, 0xCE, 0xFE, 0xDD
};

static uint8_t mtinvert[32]={
 1, 1, 1, 0, 0, 1, 1, 1,
 0, 0, 0, 0, 0, 1, 1, 1,
 1, 1, 1, 0, 0, 0, 0, 0,
 1, 1, 0, 0, 0, 0, 0, 0
};

static uint8_t domostxt=0;

static int grcolors[16]={
 COLOR_BLACK, COLOR_RED, COLOR_BLUE, COLOR_MAGENTA, 
 COLOR_GREEN, COLOR_WHITE, COLOR_BLUE, COLOR_BLUE,
 COLOR_RED, COLOR_YELLOW, COLOR_WHITE, COLOR_MAGENTA,
 COLOR_GREEN, COLOR_YELLOW, COLOR_GREEN, COLOR_WHITE
};

#define A_AII_FLASH   A_REVERSE
#define A_AII_INVERSE A_REVERSE
#define A_AII_BAR     "\xDF"

int slowkeys;

static void cputc80 (uint8_t x, uint8_t y, uint8_t value1, uint8_t value2)
{
 if ((x<40)&&(y<24)) 
 {
  if ((nogx)||(split&&(y>19)))
  {
   if (value2<0x20) mvaddch(y,x<<1,value2|0x40|A_AII_INVERSE);
    else if (value2<0x40) mvaddch(y,x<<1,value2|A_AII_INVERSE);
    else if (value2<0x80)
    {
     if ((iie->iiemode)&&(iie->mostxt))
     {
      if (domostxt && ((value2&0x40)==0x40) && (value2<0x60) && (mtapprox[value2&0x1F]))
      {
       mvaddch(y, x<<1, mtapprox[value2&0x1F]|(mtinvert[value2&0x1F]?A_AII_INVERSE:0));
      }
      else
      if (value2<0x7F) mvaddch(y,x<<1,value2|A_AII_INVERSE);
       else if (value2==0x7F) mvaddch(y,x<<1,ACS_CKBOARD|A_AII_INVERSE);
     }
     else
     {
      if (value2<0x60) mvaddch(y,x<<1,value2|A_AII_FLASH);
       else if (value2<0x80) mvaddch(y,x<<1,(value2&0x3F)|A_AII_FLASH);
     }
    }
    else if (value2<0xA0) mvaddch(y,x<<1,(value2&0x1F)|0x40);
    else if (value2==0xFF) mvaddch(y,x<<1,ACS_CKBOARD);
    else mvaddch(y,x<<1,value2&0x7F);

   if (value1<0x20) addch(value1|0x40|A_AII_INVERSE);
    else if (value1<0x40) addch(value1|A_AII_INVERSE);
    else if (value1<0x80)
    {
     if ((iie->iiemode)&&(iie->mostxt))
     {
      if (domostxt && ((value1&0x40)==0x40) && (value1<0x60) && (mtapprox[value1&0x1F]))
      {
       addch(mtapprox[value1&0x1F]|(mtinvert[value1&0x1F]?A_AII_INVERSE:0));
      }
      else
      if (value1<0x7F) addch(value1|A_AII_INVERSE);
       else if (value1==0x7F) addch(ACS_CKBOARD|A_AII_INVERSE);
     }
     else
     {
      if (value1<0x60) addch(value1|A_AII_FLASH);
       else if (value1<0x80) addch((value1&0x3F)|A_AII_FLASH);
     }
    }
    else if (value1<0xA0) addch((value1&0x1F)|0x40);
    else if (value1==0xFF) addch(ACS_CKBOARD);
    else addch(value1&0x7F);
  }
  else
  {
   if (iie->dblgrx)
   {
    if (!value2)
    {
     attron(COLOR_BLACK);
     mvprintw(y,x<<1," ");
     attroff(COLOR_BLACK);
    }
    else
    {
     value2=(((value2&0x88)>>3)|((value2&0x77)<<1));
     attron(COLOR_PAIR(value2));
     mvprintw(y,x<<1,A_AII_BAR);
     attroff(COLOR_PAIR(value2));
    }

    if (!value1)
    {
     attron(COLOR_BLACK);
     printw(" ");
     attroff(COLOR_BLACK);
    }
    else
    {
     attron(COLOR_PAIR(value1));
     printw(A_AII_BAR);
     attroff(COLOR_PAIR(value1));
    }
   }
   else
   {
    if (!value1)
    {
     attron(COLOR_BLACK);
     mvprintw(y,x<<1,"  ");
     attroff(COLOR_BLACK);
    }
    else
    {
     attron(COLOR_PAIR(value1));
     mvprintw(y,x<<1,A_AII_BAR);
     printw(A_AII_BAR);
     attroff(COLOR_PAIR(value1));
    }
   }
  }
 }
}

static void cputc40 (uint8_t x, uint8_t y, uint8_t value)
{
 if ((x<40)&&(y<24)) 
 {
  if ((nogx)||(split&&(y>19)))
  {
   if (value<0x20) mvaddch(y,x,value|0x40|A_AII_INVERSE);
    else if (value<0x40) mvaddch(y,x,value|A_AII_INVERSE);
    else if (value<0x80)
    {
     if ((iie->iiemode)&&(iie->mostxt))
     {
      if (domostxt && ((value&0x40)==0x40) && (value<0x60) && (mtapprox[value&0x1F]))
      {
       mvaddch(y, x, mtapprox[value&0x1F]|(mtinvert[value&0x1F]?A_AII_INVERSE:0));
      }
      else
      if (value<0x7F) mvaddch(y,x,value|A_AII_INVERSE);
       else if (value==0x7F) mvaddch(y,x,ACS_CKBOARD|A_AII_INVERSE);
     }
     else
     {
      if (value<0x60) mvaddch(y,x,value|A_AII_FLASH);
       else if (value<0x80) mvaddch(y,x,(value&0x3F)|A_AII_FLASH);
     }
    }
    else if (value<0xA0) mvaddch(y,x,(value&0x1F)|0x40);
    else if (value==0xFF) mvaddch(y,x,ACS_CKBOARD);
    else mvaddch(y,x,value&0x7F);
  }
  else
  {
   if (!value)
   {
    attron(COLOR_BLACK);
    mvprintw(y,x," ");
    attroff(COLOR_BLACK);
   }
   else
   {
    attron(COLOR_PAIR(value));
    mvprintw(y,x,A_AII_BAR);
    attroff(COLOR_PAIR(value));
   }
  }
 }
}

static void mark (uint16_t addr)
{
 uint8_t eights, ones, x, y, value1, value2;
 
 if (addr>0x03FF) return;
  
 value1=vramtp[page2][addr];
 if ((iie->iiemode)&&(iie->store80)) value1=vramtp[0][addr];

 value2=vramtp[2+page2][addr];
 if ((iie->iiemode)&&(iie->store80)) value2=vramtp[2][addr];
 
 eights=(addr&0x7F)/40;
 ones=(addr&0x03FF)>>7;
 x=(addr&0x7F)%40;
 y=(eights<<3)|ones;
  
 if ((iie->iiemode)&&(iie->txt80)) 
 {
  cputc80(x, y, value1, value2); 
  return;
 }
 cputc40(x, y, value1);
}

__declspec(dllexport) void ttycurses_ttywrite (uint32_t al, uint8_t v)
{
 uint16_t a, d;
 uint8_t b, c;
 
 a=al&0xFFFF;
 b=(al&65536)?2:0;
 c=(a>2047)?1:0;
 d=a&0x03FF;
 if (a<1024) return;
 if (a>3071) return;
 
 mark(d);
}

static void ttycurses_do (uint8_t a)
{
 uint16_t t;
 
 switch (a&0x000F)
 {
  case 0:
   nogx=0;
   break;
  case 1:
   nogx=1;
   break;
  case 2:
   split=0;
   break;
  case 3:
   split=1;
   break;
  case 4:
   page2=0;
   break;
  case 5:
   page2=1;
   break;
  case 6:
   hires=0;
   break;
  case 7:
   hires=1;
   break;
 }
 if (hires&&(!nogx))
 {
  mvprintw(24,75,"HGR");
 }
 else
 {
  mvprintw(24,75,"   ");
 }
 if (iie->iiemode)
 {
  iie->nogx=nogx;
  iie->split=split;
  iie->page2=page2;
  iie->hires=hires;
 }
 
 for (t=0; t<1024; t++)
 {
  mark(t);
 }
}

__declspec(dllexport) void ttycurses_ttyout (uint8_t b, uint8_t v)
{
 uint16_t a;
 
 a=0xC000|b;
 
 if ((a&0xFFF0)==0xC010)
 {
  ttycurses_ttykey &= 0x7F;
  return;
 }
 if ((a&0xFFF0)==0xC030)
 {
  
 }
 if ((a&0xFFF0)==0xC050)
 {
  ttycurses_do(a&0x000F);
  return;
 }
}

__declspec(dllexport) uint8_t ttycurses_ttyin (uint8_t b)
{
 uint16_t a;
 
 a=0xC000|b;
 if ((a&0xFFF0)==0xC000)
  return ttycurses_ttykey;
 if ((a&0xFFF0)==0xC010)
 {
  ttycurses_ttykey &= 0x7F;
  return ttycurses_ttykey;
 }
 if ((a&0xFFF0)==0xC030)
 {
  
 }
 if ((a&0xFFF0)==0xC050)
 {
  ttycurses_do(a&0x000F);
  return 0;
 }
 return 0;
}

__declspec(dllexport) void ttycurses_ttydeinit(void)
{
 if (!ttyok) return;
 clear();
 refresh();
 curs_set(1);
 endwin();
}

static uint8_t flop4080;

__declspec(dllexport) uint8_t ttycurses_ttyiie (void)
{
 uint16_t t;
 
 if (iie->iiemode)
 {
  if (flop4080!=iie->txt80)
  {
   flop4080=iie->txt80;
   
   if (iie->txt80)
   {
    mvprintw(24,72,"80");
   }
   else
   {
    mvprintw(24,72,"40");
    for (t=0; t<24; t++)
    {
     move(t,40);
     clrtoeol();
    }
   }
  }
 }
 

 for (t=0; t<1024; t++)
 {
  mark(t);
 }
}

__declspec(dllexport) void ttycurses_ttyinit(IIE *main_iie)
{
 int t;
 
 flop4080=0;
 slowkeys=0;

 iie=main_iie;
 
 vramtp[0]=&((iie->ram)[0x00400]);
 vramtp[1]=&((iie->ram)[0x00800]);
 vramtp[2]=&((iie->ram)[0x10400]);
 vramtp[3]=&((iie->ram)[0x10800]);
 
 if (iie->iiemode)
 {
  iie->alt[0]=iie->alt[1]=0;
  domostxt=iie->enh;
 }
 
 initscr();
 raw();
 keypad(stdscr, TRUE);
 noecho();
 curs_set(0);
 nodelay(stdscr, TRUE);
 start_color();
 use_default_colors();
 for (t=0; t<256; t++)
 {
  init_pair(t, grcolors[t&0x0F], grcolors[(t&0xF0)>>4]);
 }
 ttyok=1;

 if (iie->iiemode) mvprintw(24,72,"40");
}

void twink (int foo)
{
 Beep(foo?500:1000,100);
}

__declspec(dllexport) int ttycurses_ttyloop(void)
{
 int k;

 refresh();
 
 if (slowkeys && (ttycurses_ttykey&0x80)) return 0;
 
 k=getch();
 if (iie->iiemode)
 {
  if (k==KEY_F(37)) 
  {
   iie->alt[0]=(iie->alt[0])^0x80; 
   twink(iie->alt[0]);
   if (iie->alt[0]) mvprintw(24,3,"OA");
              else  mvprintw(24,3,"  ");
  }
  if (k==KEY_F(38)) 
  {
   iie->alt[1]=(iie->alt[1])^0x80; 
   twink(iie->alt[1]);
   if (iie->alt[1]) mvprintw(24,6,"SA");
              else  mvprintw(24,6,"  ");
  }
  if (k==KEY_F(39))
  {
   slowkeys=!slowkeys;
   twink(slowkeys);
   if (slowkeys) mvprintw(24,9,"SK");
           else  mvprintw(24,9,"  ");
  }
 }
 if (k!=ERR)
 {
  switch (k)
  {
   case KEY_BACKSPACE:
    ttycurses_ttykey=0xFF;
    return 0;
   case KEY_UP:
    ttycurses_ttykey=0x8B;
    return 0;
   case KEY_DOWN:
    ttycurses_ttykey=0x8A;
    return 0;
   case KEY_LEFT:
    ttycurses_ttykey=0x88;
    return 0;
   case KEY_RIGHT:
    ttycurses_ttykey=0x95;
    return 0;
   case '\n':
    ttycurses_ttykey=0x8D;
    return 0;

    /* Slot config */
   case KEY_F(1): return 11;
   case KEY_F(2): return 12;
   case KEY_F(3): return 13;
   case KEY_F(4): return 14;
   case KEY_F(5): return 15;
   case KEY_F(6): return 16;
   case KEY_F(7): return 17;
   case KEY_F(8): return 10;

   case KEY_F(10):
   case KEY_F(40):
    return -1;
   case KEY_F(36):
   case KEY_BREAK:
    return 1;
   case KEY_F(41):
    return 2;
   default:
    if (k<0x80)
    {
     /* flip case */
     if ((k>='A')&&(k<='Z')) k|=32;
      else if ((k>='a')&&(k<='z')) k&=223;
     ttycurses_ttykey=k|0x80;
    }
  }
 }
 return 0;
}
